// Copyright (C) 2011 INRIA Sophia Antipolis Méditerranée
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

#include "ReverseNetworkTable.h"
#include <omnetpp.h>
#include "Synapse.h"

namespace oversim
{

Define_Module(ReverseNetworkTable);

void ReverseNetworkTable::initialize(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces
    // are registered, address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY)
        return;
    
    synapsePort = par("port");
    
    std::vector<std::string> availableOverlayNetIDs =
		cStringTokenizer(par("availableOverlayNetIDs"), " ").asVector();
		
	for (std::vector<std::string>::iterator itNetID = availableOverlayNetIDs.begin();
		itNetID != availableOverlayNetIDs.end(); ++itNetID) {
		SynapseNodeSet *newSet = new SynapseNodeSet();
		WATCH_SET(*newSet);
		nodeMap[OverlayKey(*itNetID).toString()] = newSet;
	}
    WATCH_MAP(nodeMap); 
    WATCH_MAP(sortedNodes); 
}

void ReverseNetworkTable::handleMessage(cMessage* msg)
{
    //error("this module doesn't handle messages, it runs only in initialize()");
}

ReverseNetworkTable::~ReverseNetworkTable(){
	
	for (SynapseNodeMap::iterator it = nodeMap.begin();
		it != nodeMap.end(); ++it)
		delete it->second;
	
	for (NodeIteratorMap::iterator it = sortedNodes.begin();
		it != sortedNodes.end(); ++it)
		delete it->second;
}


void ReverseNetworkTable::addNode(NodeHandle& node)
{
	OverlayKey primaryNetID = node.getNetID();
	std::vector<std::string> secondaryNetIDs = node.getSecondaryNetIDs();
	
	if (!primaryNetID.isUnspecified()) {
		addNode(node, primaryNetID.toString(16));
	}
	
	for (std::vector<std::string>::iterator it = secondaryNetIDs.begin();
		it != secondaryNetIDs.end(); ++it) {
		if (((std::string) *it).length() > 0)
			addNode(node, *it);
	}
}
	
void ReverseNetworkTable::addNode(TransportAddress& node, NetworkIdRef netID)
{
	SynapseNodeSet *nodeSet;
	NodeIteratorVector *nodeVector;
	
	//get the correspondent nodeset for the give networkID
	nodeSet = getNodeSetForNetID(netID, true);
	//get the correspondent nodevector for the give networkID
	nodeVector = getNodeVectorForNetID(netID, true);
	
	TransportAddress insertNode = TransportAddress(node);
	insertNode.setPort(synapsePort);
	
	std::pair<SynapseNodeSet::iterator, bool> res = nodeSet->insert(insertNode);
	//if succesfully added, add the iterator in the nodevector.
	if (res.second)
		nodeVector->push_back(res.first);
}

void ReverseNetworkTable::removeNode(NodeHandle& node)
{
	OverlayKey primaryNetID = node.getNetID();
	std::vector<std::string> secondaryNetIDs = node.getSecondaryNetIDs();
	
	if (!primaryNetID.isUnspecified()) {
		removeNode(node, primaryNetID.toString(16));
	}
	
	for (std::vector<std::string>::iterator it = secondaryNetIDs.begin();
		it != secondaryNetIDs.end(); ++it) {
		removeNode(node, *it);
	}
}

void ReverseNetworkTable::removeNode(TransportAddress& node, NetworkIdRef netID) {
	SynapseNodeSet *nodeSet;
	NodeIteratorVector *nodeVector;
	
	//get the correspondent nodeset for the give networkID
	nodeSet = getNodeSetForNetID(netID, false);
	//get the correspondent nodevector for the give networkID
	nodeVector = getNodeVectorForNetID(netID, false);
	
	if (nodeSet != NULL) {
		TransportAddress remNode = TransportAddress(node);
		remNode.setPort(synapsePort);
		SynapseNodeSet::iterator res = nodeSet->find(remNode);
		//if succesfully added, add the iterator in the nodevector.
		if (res != nodeSet->end()) {
			nodeSet->erase(*res);
			
			if (nodeVector != NULL) {
				// TODO: optimize!
				for (NodeIteratorVector::iterator itV = nodeVector->begin();
					itV != nodeVector->end(); ++itV)
					if (*itV == res) nodeVector->erase(itV);
			}
		}
	}
}

const TransportAddress& ReverseNetworkTable::getRandomNodeForNetwork(NetworkIdRef netID)
{
	NodeIteratorVector *nodeVector = getNodeVectorForNetID(netID, false);
	
	if (nodeVector == NULL)
		return NodeHandle::UNSPECIFIED_NODE;
	
	int rndIndex = intuniform(1, nodeVector->size());
	
	NodeIteratorVector::iterator itV = nodeVector->begin();
	for (int i=0; (i<rndIndex) && (itV != nodeVector->end()); i++, ++itV);
	
	if (itV != nodeVector->end())
		return ((const NodeHandle&) **itV);
	else
		return NodeHandle::UNSPECIFIED_NODE;
}

void ReverseNetworkTable::refreshNode(NodeHandle& node)
{
	removeNode(node);
	addNode(node);
}

SynapseNodeSet *ReverseNetworkTable::getNodeSetForNetID(NetworkIdRef netID, bool createIfNonExistent)
{
	SynapseNodeMap::iterator itMap = nodeMap.find(netID);
	SynapseNodeSet *nodeSet = NULL;
	
	if (itMap == nodeMap.end()) { //set not present
		if (createIfNonExistent) {
			nodeSet = new SynapseNodeSet();
			WATCH_SET(*nodeSet);
			nodeMap[netID] = nodeSet;
		}
	} else {
		nodeSet = itMap->second;
	}
	
	return nodeSet;
}
	
NodeIteratorVector *ReverseNetworkTable::getNodeVectorForNetID(NetworkIdRef netID, bool createIfNonExistent) 
{
	NodeIteratorMap::iterator itIter = sortedNodes.find(netID);
	NodeIteratorVector *nodeVector = NULL;
	
	if (itIter == sortedNodes.end()) {
		if (createIfNonExistent) {
			nodeVector = new NodeIteratorVector();
			//WATCH_PTRVECTOR(*nodeVector);
			sortedNodes[netID] = nodeVector;
		}
	} else {
		nodeVector = itIter->second;
	}
	
	return nodeVector;
};
}//namespace