//
// Copyright (C) 2011 INRIA Sophia Antipolis Mediterranee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SynapseController.cc
 * @author Vincenzo Ciancaglini
 */

#include "SynapseController.h"
#include <../underlay/simpleunderlay/SimpleUDP.h>
#include <GlobalStatistics.h>
#include <GlobalNodeList.h>
#include "ISynapseBaseOverlay.h"
#include "BaseOverlay.h"
#include <CommonMessages_m.h>
#include "SynapseMessages_m.h"


namespace oversim {
	
	Define_Module(SynapseController);
	
	SynapseController::SynapseController() 
	{}
	
	void SynapseController::initConnectedOverlaysList()
	{
		// Retrieve the configuration parameters
		std::vector<std::string> connectedOverlayNames;
		cStringTokenizer(par("connectedOverlays"), " ").asVector();

		if (connectedOverlays.size() > availableOverlaysByNetID.size())
		{
			throw cRuntimeError(std::string("SynapseController::"
						"initConnectedOverlaysList(): there are more connected overlays declared than available ones. Check configuration.").c_str());
		}

		std::vector<OverlayKey> netIDList;

		for (std::vector<std::string>::iterator itNetName = connectedOverlayNames.begin(); 
				itNetName != connectedOverlayNames.end(); ++itNetName) 
		{
			ConnectedOverlaysMap::iterator itTest = connectedOverlays.find(*itNetName);

			// TODO: check that the error is thrown
			// If the netID is already in the map, it means that it has been declared twice in the config list.
			if (itTest != connectedOverlays.end())
				throw cRuntimeError(std::string("SynapseController::"
							"initConnectedOverlaysList(): duplicate entries in connectedOverlays.").c_str());
			else {
				OverlaysParametersMap::iterator itOv = availableOverlaysByName.find(*itNetName);

				if (itOv == availableOverlaysByName.end()) 
					throw cRuntimeError(std::string("SynapseController::"
								"initConnectedOverlaysList(): connected overlay name non existent. Declare it first in the available overlay list.").c_str());

				const char *ovNetID = (const char *) ((* itOv->second)[OVERLAY_NETID]->c_str());
				const char *ovName = (const char *) ((* itOv->second)[OVERLAY_NAME]->c_str());
				const char *ovType = (const char *) ((* itOv->second)[OVERLAY_TYPE]->c_str());
				const char *ovSyn = (const char *) ((* itOv->second)[OVERLAY_SYN_MODULE]->c_str());

				//std::cout << "NID: " << ovNetID << "\nName: " << ovName << "\nType: " << ovType << "\nMod: " << ovSyn;
				cModule *ovModule = createSubOverlay(ovNetID, ovName, ovType, ovSyn);

				connectedOverlays[OverlayParameter(ovNetID)] = ovModule;
				connectedOverlaysByName[OverlayParameter(ovName)] = ovModule;

				netIDList.push_back(OverlayKey(ovNetID));
			}
		}

		//update all the nodehandles with all the secondary netID.
		for (std::vector<OverlayKey>::iterator itC = netIDList.begin();
				itC != netIDList.end(); ++itC) {
			thisNode.addSecondaryNetID(*itC);
			for (ConnectedOverlaysMap::const_iterator itC2 = connectedOverlays.begin();
					itC2 != connectedOverlays.end(); ++itC2) {
				ISynapseBaseOverlay * bo = check_and_cast<ISynapseBaseOverlay *> (itC2->second);
				bo->addSecondaryNetworkID(*itC);
			}
		}
	}
	
	void SynapseController::initOverlaysMap() 
	{
		// Retrieve the configuration parameters
		std::vector<std::string> availableOverlayNames =
		cStringTokenizer(par("availableOverlayNames"), " ").asVector();
		
		std::vector<std::string> availableOverlayTypes =
		cStringTokenizer(par("availableOverlayTypes"), " ").asVector();
		
		std::vector<std::string> availableOverlayNetIDs =
		cStringTokenizer(par("availableOverlayNetIDs"), " ").asVector();
		
		std::vector<std::string> availableOverlaySynModules =
		cStringTokenizer(par("availableOverlaySynapseModules"), " ").asVector();
		
		// TODO: check that the error is thrown
		// If the 3 lists have different size, we throw an error (they must all have
		// the same number of elements
		unsigned int aotSize = availableOverlayTypes.size();
		
		if ((aotSize != availableOverlayNetIDs.size()) ||
		    (aotSize != availableOverlayNames.size())  ||
		    (aotSize != availableOverlaySynModules.size()))
		{
			throw cRuntimeError(std::string("SynapseController::"
			"initOverlaysMap(): availableOverlayNames, availableOverlayTypes and availableOverlayNetIDs "
			"have different length. Check your config file.").c_str());
		}
		
		//Create the availableOverlaysMap, with key=NetIDRef and value=OverlayParameterList
		for (std::vector<std::string>::iterator itNetID = availableOverlayNetIDs.begin(),
			itType = availableOverlayTypes.begin(),
			itName = availableOverlayNames.begin(),
			itSynMod = availableOverlaySynModules.begin(); 
			itNetID != availableOverlayNetIDs.end(); ++itNetID,++itType,++itName, ++itSynMod) 
		{
			OverlaysParametersMap::iterator itTest = availableOverlaysByNetID.find(*itNetID);
		
		// TODO: check that the error is thrown
		// If the netID is already in the map, it means that it has been declared twice in the config list.
			if (itTest != availableOverlaysByNetID.end())
				throw cRuntimeError(std::string("SynapseController::"
				"initOverlaysMap(): duplicate entries in availableOverlayNetIDs, use different values for network IDs").c_str());
			else 
			{
				OverlayParametersList *opl = new OverlayParametersList();
 				opl->push_back(new OverlayParameter(*itType)); // new GIVES OUT POINTERS
 				opl->push_back(new OverlayParameter(*itName));
 				opl->push_back(new OverlayParameter(*itNetID));
 				opl->push_back(new OverlayParameter(*itSynMod));
				
				availableOverlaysByNetID[OverlayParameter(*itNetID)] = opl;
				availableOverlaysByName[OverlayParameter(*itName)] = opl;
			}
		}	
	}
	
	cModule * SynapseController::createSubOverlay(const char *netID, const char *overlayName, const char *overlayType, const char *synapseModuleName) {
		// Get the specified module type
		cModuleType* moduleType = cModuleType::get ( overlayType );
		//const char *ovlName = "subOverlay"; //da rendere generico
		
		// Create the overlay compound module
		//cModule* tempModule = moduleType->create ( ovlName, getParentModule(), index, maxOverlays ); //TODO levami
		
		// The following would create an overlay named "synapsecontroller.ovlerayName"
		//std::string completeName =  std::string(getName()) + "." + std::string(overlayName);
		// Instead we use just overlayName (I keep the above line in case we actually need to nest the namespace)
		//
		// getParentModule()[SynapseModules(overlay)]->getParentModule()[[*]overlayHost]
		// getParentModule()->getParentModule() because we need to attach it to the overlayHost, in order to retrieve
		// the BootstrapList and other data structures.
		cModule* newOverlayModule = moduleType->create (overlayName, getParentModule()->getParentModule() );
		
		//Connect the gates, to the controller and to UDP
		// getParentModule()[SynapseModules(overlay)]->getParentModule()[[*]overlayHost]
		cModule *udp = check_and_cast<SimpleUDP *>(getParentModule()->getParentModule()->getSubmodule("udp", 0));
		//Connect the overlay's appIn/Out to the synapseController
		newOverlayModule->gate("appOut")->connectTo(getOrCreateFirstUnconnectedGate("ovlAppIn", 0, false, true));
		getOrCreateFirstUnconnectedGate("ovlAppOut",0, false, true)->connectTo(newOverlayModule->gate("appIn"));
		//Connect the overlay's updIn/Out directly to Udp module. 
		newOverlayModule->gate("udpOut")->connectTo(udp->getOrCreateFirstUnconnectedGate("appIn", 0, false, true));
		udp->getOrCreateFirstUnconnectedGate("appOut", 0, false, true)->connectTo(newOverlayModule->gate("udpIn"));
		
		newOverlayModule->finalizeParameters();
		// Build the module internals
		newOverlayModule->buildInside();
		newOverlayModule->scheduleStart(simTime());
		
		// Set the networkID for the module
		// Get the internal module implementing the SynapseBaseOverlay interface
		cModule *ovlSynBase = newOverlayModule->getSubmodule ( synapseModuleName );
		ISynapseBaseOverlay *ovlNode = check_and_cast<ISynapseBaseOverlay *> ( ovlSynBase );
		//SynapseBaseOverlay *ovlSynNode = check_and_cast<SynapseBaseOverlay *> ( ovlSynBase );
		ovlNode->setNetworkID ( OverlayKey(netID) );
		
		// THE IP ADRESS IS SET IN THE INIT PHASE FOR ALL OVERLAYS (TAKEN FROM THE PARENT NODE)
		// Set the same IP adress as the synapse controller (they are supposed to be on the same host)
		//BaseOverlay *ovlNodeBase = check_and_cast<BaseOverlay *>(ovlSynBase);
		//IPvXAddress ip = getThisNode().getIp();
		//TransportAddress *childNode = ((TransportAddress *) &(ovlNodeBase->getThisNode()));
		//childNode->setIp(ip);
		
		// TODO: LEVAMI DEBUG;
		std::cout << "Created subOverlay " << overlayName << /*", ID: " << ((NodeHandle *) childNode)->getKey() <<*/ ", netID: " << netID << endl;
		//return newOverlayModule;
		return ovlSynBase;
	}
	
	void SynapseController::initializeOverlay(int stage) 
	{
		Enter_Method_Silent();
		
		//fetch some parameters
		synapseDebug = par("synapseDebug");
		
		//get the friends module
		reverseNetworkTable = check_and_cast<ReverseNetworkTable*>
                  (getParentModule()->getSubmodule("reverseNetworkTable"));
			   
		requestTable = check_and_cast<SynapseRoutingTable*>
                  (getParentModule()->getSubmodule("requestTable"));
		
		// init the overlay submodules
		if (stage == MIN_STAGE_OVERLAY) {
			initOverlaysMap();
			initConnectedOverlaysList();
			
			// Synapse provides KBR services
			kbr = true;
		}
		// once they are initialized, register yourself as component 
		if (stage == MAX_STAGE_OVERLAY) {
			// After all the suboverlays have been initiated
			// we need to re-register this module as THE TRUE AND ONLY overlay,
			// otherwise all the broadcasts will go to the last suboverlay initialized.
			registerComp(getThisCompType(), this);
			
			ConnectedOverlaysMap::iterator itOv;
			for (itOv = connectedOverlays.begin();
				itOv != connectedOverlays.end(); ++itOv) {
				BaseOverlay *ovlMod = check_and_cast<BaseOverlay *>(itOv->second);
				//run some post init commands on all overlays
				ovlMod->postInitializeOverlay();
				//register the synapseController to all the overlays
				ovlMod->registerComp(CONTROLLER_COMP, this);
			}
		}
	}
	
	void SynapseController::finishOverlay() 
	{
		// TODO: delete all tables and maps.
		// Collect statistics.
		// DA FARE DA FARE DA FARE!!!
	}
	
	// Process messages passed up from the overlay.
	void SynapseController::handleMessage(cMessage* msg)
	{
		//timers and timeouts
		if (internalHandleMessage(msg)) {
			return;
		}
		// message coming from one of the subOverlays
		if (msg->arrivedOn("ovlAppIn") /*||
			msg->arrivedOn("direct_in")*/) {
			// common API from the suboverlay
			CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
			if (commonAPIMsg != NULL) {
				handleCommonAPILowerMessage(commonAPIMsg);
			} else /*if (msg->arrivedOn("ovlAppIn"))*/ {
				// TODO: What kind of messages to we want to measure here?
				cPacket* packet = check_and_cast<cPacket*>(msg);
				//RECORD_STATS(numOverlayReceived++;
				//			bytesOverlayReceived += packet->getByteLength());
				handleLowerMessage(msg); // other message (is there any other?)
			}/*else delete msg;*/}
		// message either coming from one of the components (READY state)
		// or from the app sending a ROUTE/FORWARD command to be passed down
		else if (msg->arrivedOn("direct_in")) {
			CompReadyMessage* readyMsg = dynamic_cast<CompReadyMessage*>(msg);
			if (readyMsg != NULL) {
				handleReadyMessage(readyMsg);
				return;
			}
			CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
			if (commonAPIMsg != NULL)
				handleCommonAPIHigherMessage(commonAPIMsg);
		} // other messages from app (are there any?)
		else if (msg->arrivedOn("appIn"))
			handleUpperMessage(msg);
		else if (msg->arrivedOn("udpIn")) {
			cPacket* packet = check_and_cast<cPacket*>(msg);
			//RECORD_STATS(numUdpReceived++; bytesUdpReceived += packet->getByteLength());
			// debug message
			if (debugOutput && !ev.isDisabled()) {
				UDPControlInfo* udpControlInfo =
					check_and_cast<UDPControlInfo*>(msg->getControlInfo());
				EV << "[BaseApp:handleMessage() @ " << thisNode.getIp()
				<< " (" << overlay->getThisNode().getKey().toString(16) << ")]\n"
				<< "    Received " << *msg << " from "
				<< udpControlInfo->getSrcAddr() << endl;
			}
			handleUDPMessage(msg);
		} else if(msg->arrivedOn("tcpIn")) {
			handleTCPMessage(msg);
		} else if (msg->arrivedOn("trace_in")) {
			handleTraceMessage(msg);
		}else {
			delete msg;
		}
	}
	
	void SynapseController::handleCommonAPILowerMessage(CommonAPIMessage* commonAPIMsg)
	{
		cPacket* tempMsg = commonAPIMsg->decapsulate();
		NodeHandle comp;
		// process interface control information
		OverlayCtrlInfo* overlayCtrlInfo =
		dynamic_cast<OverlayCtrlInfo*>(commonAPIMsg->removeControlInfo());
		
		if (overlayCtrlInfo != NULL) {
			tempMsg->setControlInfo(overlayCtrlInfo);
		}
		
		switch (commonAPIMsg->getType()) {
			case KBR_DELIVER:
			{
				KBRdeliver* apiMsg = dynamic_cast<KBRdeliver*>(commonAPIMsg);
				OverlayKey key = apiMsg->getDestKey();
				// TODO: MANAGE MULTIOVERLAY (got away for now...)
				NodeHandle nextHopNode = getThisNode(); //overlay->
				
				//first call forward, then deliver
				forward(&key, &tempMsg, &nextHopNode);
				
				if(tempMsg != NULL) {
					//if key or nextHopNode is changed send msg back to overlay
					if ((!key.isUnspecified() && key != apiMsg->getDestKey()) ||
						(!nextHopNode.isUnspecified()
						&& nextHopNode != overlay->getThisNode())) {
						forwardResponse(key, tempMsg, nextHopNode);
						}
						else {
							//RECORD_STATS(numOverlayReceived++;
							//bytesOverlayReceived += tempMsg->getByteLength());
							
							assert(overlayCtrlInfo->getTransportType()
							== ROUTE_TRANSPORT);
							
							// debug message
							if (debugOutput) {
								EV << "[BaseApp:handleCommonAPIMessage() @ "
								<< thisNode.getIp() << " ("
								<< overlay->getThisNode().getKey().toString(16) << ")]\n"
								<< "    Received " << *tempMsg << " from "
								<< overlayCtrlInfo->getSrcRoute() << endl;
							}
							
							//handle RPC first
							BaseRpcMessage* rpcMessage
							= dynamic_cast<BaseRpcMessage*>(tempMsg);
							if (rpcMessage!=NULL) {
								internalHandleRpcMessage(rpcMessage);
							} else {
								deliver(apiMsg->getDestKey(), tempMsg);
							}
						}
				}
				break;
			}
			
			case KBR_FORWARD:
			{
				KBRforward* apiMsg = dynamic_cast<KBRforward*>(commonAPIMsg);
				OverlayKey key = apiMsg->getDestKey();
				NodeHandle nextHopNode = apiMsg->getNextHopNode();
				
				forward(&key, &tempMsg, &nextHopNode);
				
				//if message ist not deleted send it back
				if(tempMsg != NULL) {
					if(nextHopNode == apiMsg->getNextHopNode())
						//do we need this?
					nextHopNode = NodeHandle::UNSPECIFIED_NODE;
					forwardResponse(key, tempMsg, nextHopNode);
				}
				break;
			}
			
			case KBR_UPDATE:
			{
				KBRupdate* apiMsg = dynamic_cast<KBRupdate*>(commonAPIMsg);
				update(apiMsg->getNode(), apiMsg->getJoined());
				
				break;
			}
			
			case KBR_NOTIFY:
			{
				KBRnotify *notifyMsg = dynamic_cast<KBRnotify *>(commonAPIMsg);
				comp = notifyMsg->getSrcCompNode();
				handleForwardNotification(tempMsg->dup(), comp);
				break;
			}
			default:
			{
				break;
			}
		}
		delete commonAPIMsg;
		delete tempMsg;
	}
	
	void SynapseController::handleForwardNotification(cPacket* msg, NodeHandle comp)
	{
		//std::cout << "ABBIAMO LA NOTIFICA" << endl;
		NodeHandle src;
		OverlayKey routeKey;
		BaseRouteMessage *routeMsg;
		BaseAppDataMessage *appMsg;
		BaseRpcMessage *rpcMsg;
		
		OverlayCtrlInfo* overlayCtrlInfo =
		dynamic_cast<OverlayCtrlInfo*>(msg->getControlInfo());
		
		OverlayKey key;
		FindNodeCall *findCall;
		
		BaseOverlayMessage *encMsg = dynamic_cast<BaseOverlayMessage *>(msg);
		switch (encMsg->getType())
		{
			//These 2 we can ignore
			case OVERLAYSIGNALING:
				if (synapseDebug)
					std::cout << "SYNAPSECONTROLLER::handleForwardNotification\n\t" 
							<< "Overlay signaling message." << endl;
				break;
			case APPDATA:
				//ctrl = check_and_cast<OverlayCtrlInfo*>(encMsg->getControlInfo());
				appMsg = dynamic_cast<BaseAppDataMessage *>(encMsg);
				if (synapseDebug) {
					if (overlayCtrlInfo) {
						src = overlayCtrlInfo->getSrcNode();
						//TODO LEVAMI DEBUG
						std::cout << "SYNAPSECONTROLLER::handleForwardNotification\n\t" 
								<< "Overlay: " << comp.getNetID().toString(16) 
								<< ", Node: " << comp.getIp().str() << ":" << comp.getPort() 
								<< "(" << comp.getKey().toString(16) << ")\n"
								<< "\tis handling an APPDATA message \n\tfrom Node"
								<< src.getIp().str() << ":" << src.getPort() 
								<< "(" << src.getKey().toString(16) << ")"
								<< endl;
					} else {
						//TODO LEVAMI DEBUG
						std::cout << "SYNAPSECONTROLLER::handleForwardNotification\n\t" 
								<< "Overlay: " << comp.getNetID().toString(16) 
								<< ", Node: " << comp.getIp().str() << ":" << comp.getPort() 
								<< "(" << comp.getKey().toString(16) << ")\n"
								<< "\tis handling an APPDATA message \n\tfrom Node Unknown."
								<< endl;
					}
				}
				break;
			//these 2 we need to consider to scrape out inportant informations
			case OVERLAYROUTE:
				routeMsg = dynamic_cast<BaseRouteMessage *>(encMsg);
				src = NodeHandle(routeMsg->getSrcNode());
				routeKey = routeMsg->getDestKey();
				
				if (synapseDebug)
					std::cout << "SYNAPSECONTROLLER::handleForwardNotification\n\t" 
							<< "Overlay: " << comp.getNetID().toString(16) 
							<< ", Node: " << comp.getIp().str() << ":" << comp.getPort() 
							<< " (" << comp.getKey().toString(16) << ")\n"
							<< "\tis routing an OVERLAYROUTING msg to key " << routeKey.toString(16)
							<< "\n\tfrom Node: " << src.getIp().str() << ":" << src.getPort()
							<< " (" << src.getKey().toString(16) << ")\n\t"
							<< "Secondary overlays: " << (int) src.getSecondaryNetIDs().size()
							<< endl;
						
				// we could also analyze the overlayControl, but for now 
				// we keep it simple and universal
				// if the source node is not a legacy node
				if (!isLegacyNode(src)) {
					// we add the node as synapse for all the overlays it is 
					// connected to
					reverseNetworkTable->addNode(src);
					src.setPort(thisNode.getPort());
					sendSynapseOffer(src, routeKey, comp.getNetID());
				}
				break;
			case RPC:
				//TODO LEVAMI DEBUG
				rpcMsg = dynamic_cast<BaseRpcMessage *>(encMsg);
				findCall = dynamic_cast<FindNodeCall *>(rpcMsg);
				key = OverlayKey::UNSPECIFIED_KEY;
				if (findCall != NULL)
					key = findCall->getLookupKey();
				if (synapseDebug)
					std::cout << "SYNAPSECONTROLLER::handleForwardNotification:"
							<< thisNode.getIp() << ":" << thisNode.getPort() <<"\n\t"
							<< "RPC Call message from: " << rpcMsg->getSrcNode().getIp().str()
							<< ":" << rpcMsg->getSrcNode().getPort() << "\n\t"
							<< "Secondary overlays: " << (int) rpcMsg->getSrcNode().getSecondaryNetIDs().size()
							<< endl;
				src = NodeHandle(rpcMsg->getSrcNode());
				// we could also analyze the overlayControl, but for now 
				// we keep it simple and universal
				// if the source node is not a legacy node
				if (!isLegacyNode(src)) {
					// we add the node as synapse for all the overlays it is 
					// connected to
					reverseNetworkTable->addNode(src);
					src.setPort(thisNode.getPort());
					sendSynapseOffer(src, key, comp.getNetID());
				}
				break;
			default:
				break;
		}
		
		delete msg;
	}
	
	void SynapseController::sendSynapseOffer(NodeHandle& dest, OverlayKey key, OverlayKey contextNetID)
	{
		SynapseOfferMessage *synOffMsg = new SynapseOfferMessage();
		
		synOffMsg->setSrcNode(thisNode);
		synOffMsg->setDestKey(key);
		synOffMsg->setOriginalNetID(contextNetID);
		
		if (synapseDebug)
			std::cout << "SYNAPSECONTROLLER::sendSyapseOffer: " 
					<< thisNode.getIp() << ":" << thisNode.getPort() <<"\n\t"
					<< "sending offer associated to key: " << key.toString()
					<< " for overlay: " << contextNetID.toString()
					<< "\n\tto node: " << dest.getIp() << ":" << dest.getPort()
					<< "\n\tSecondary overlays: " << (int) thisNode.getSecondaryNetIDs().size()
					<< endl;
					
		sendMessageToUDP(dest, synOffMsg);
	}
	
	bool SynapseController::isLegacyNode(NodeHandle node){
		return node.isNetIDUnspecified();
	}
	
     void SynapseController::handleCommonAPIHigherMessage(CommonAPIMessage* commonAPIMsg)
	{
		// PER ORA FORWARDIAMO TUTTO A TUTTI
		// TODO: decide multiforward strategy
		for (ConnectedOverlaysMap::iterator itOv = connectedOverlays.begin();
			itOv != connectedOverlays.end(); ++itOv)
			sendDirect((cMessage *) commonAPIMsg->dup(), itOv->second->gate("direct_in"));
		delete commonAPIMsg;
	}
    
	void SynapseController::handleLowerMessage(cMessage* msg)
	{
		delete msg;
	}
	
	void SynapseController::handleUpperMessage(cMessage* msg)
	{
		delete msg;
	}
	
	void SynapseController::handleUDPMessage(cMessage* msg)
	{
		BaseSynapseMessage *synMsg;
		
		if ((synMsg = dynamic_cast<SynapseOfferMessage *>(msg)) != NULL) {
			SynapseOfferMessage *synOffMsg = (SynapseOfferMessage *) synMsg;
			if (synapseDebug)	
				std::cout << "SYNAPSECONTROLLER::handleUDPMessage: "
						<< thisNode.getIp() << ":" << thisNode.getPort() <<"\n\t"
						<< "received SynapseOffer from Node: " 
						<< synOffMsg->getSrcNode().getIp() << ":" << synOffMsg->getSrcNode().getPort() 
						<< "\n\tAssociated KEY: " << synOffMsg->getDestKey().toString()
						<< "\n\tSecondary overlays: " << (int) synOffMsg->getSrcNode().getSecondaryNetIDs().size()
						<< endl;
			handleSynapseOfferMessage(synOffMsg);
		} 
		else if ((synMsg = dynamic_cast<SynapseRequestMessage *>(msg)) != NULL) {
			SynapseRequestMessage *synReqMsg = (SynapseRequestMessage *) synMsg;
			// TODO call thre asynchronous synapseRequest handler
			handleSynapseRouteRequestMessage(synReqMsg);
		} 
		else if ((synMsg = dynamic_cast<SynapseResponseMessage *>(msg)) != NULL) {
			SynapseResponseMessage *synRespMsg = (SynapseResponseMessage *) synMsg;
			// TODO call thre asynchronous synapseResponse handler
			handleSynapseRouteResponseMessage(synRespMsg);
		}
		else
			if (synapseDebug)
				std::cout << "SYNAPSECONTROLLER::handleUDPMessage: " << "Other received...weird" << endl;
		delete msg;
	}
	
	bool SynapseController::handleRpcCall(BaseCallMessage* msg)
	{
		if (state != READY) {
			EV << "[SynapseController::handleRpcCall() @ " << thisNode.getIp()
			<< " (" << thisNode.getKey().toString(16) << ")]\n"
			<< "    Received RPC call and state != READY"
			<< endl;
			return false;
		}
		
		// delegate messages
		RPC_SWITCH_START( msg )
		// RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
		RPC_DELEGATE( SynapseRoute, handleSynapseRouteCall );
		RPC_SWITCH_END( )
		
		return RPC_HANDLED;
	}
	
	void SynapseController::handleRpcResponse(BaseResponseMessage* msg,
									  cPolymorphic* context, int rpcId,
								   simtime_t rtt)
	{
		RPC_SWITCH_START(msg)
		RPC_ON_RESPONSE( SynapseRoute ) {
			handleSynapseRouteResponse(_SynapseRouteResponse);
			EV << "[SynapseController::handleRpcResponse() @ " << thisNode.getIp()
			<< " (" << thisNode.getKey().toString(16) << ")]\n"
			<< "    Received a Join RPC Response: id=" << rpcId << "\n"
			<< "    msg=" << *_SynapseRouteResponse << " rtt=" << rtt
			<< endl;
			break;
		}
		//TODO add Join/Invite...
		RPC_SWITCH_END( )
	}
	
	void SynapseController::handleRpcTimeout(BaseCallMessage* msg,
									 const TransportAddress& dest,
								  cPolymorphic* context, int rpcId,
								  const OverlayKey&)
	{
		RPC_SWITCH_START(msg)
		RPC_ON_CALL( SynapseRequest ) {
			EV << "[SynapseController::handleRpcTimeout() @ " << thisNode.getIp()
			<< " (" << thisNode.getKey().toString(16) << ")]\n"
			<< "    FindNode RPC Call timed out: id=" << rpcId << "\n"
			<< "    msg=" << *_SynapseRequestCall
			<< endl;
			handleSynapseRouteTimeout();
			break;
		}
		//TODO add Join/Invite...
		RPC_SWITCH_END( )
	}

	void SynapseController::handleSynapseOfferMessage(SynapseOfferMessage *synOffMsg) 
	{
		NodeHandle src = synOffMsg->getSrcNode();
		OverlayKey key = synOffMsg->getDestKey();
			
		reverseNetworkTable->refreshNode(src);
		// TODO implement
		//if (opportunisticRouting && key && isGoodDeal(src, key))
		// sendSynapseRequestCall();
	}
	
	void SynapseController::handleSynapseRequestMessage(SynapseRequestMessage *synReqMsg) 
	{
		delete synReqMsg;
	}
	
	void SynapseController::handleSynapseResponseMessage(SynapseResponseMessage *synRespMsg) 
	{
		delete synRespMsg;
	}
	
	void SynapseController::handleSynapseRouteCall(SynapseRequestCall *synReqCall) 
	{
	}
	
	void SynapseController::handleSynapseRouteResponse(SynapseResponseCall *synRespCall) 
	{
	}
	
	void SynapseController::handleSynapseRouteTimeout() 
	{
	}
	
	void SynapseController::handleTCPMessage(cMessage* msg)
	{
		delete msg;
	}
	
	void SynapseController::handleReadyMessage(CompReadyMessage* readyMsg)
	{
		if (((bool)par("joinOnApplicationRequest") == false) &&
			readyMsg->getReady() &&
			readyMsg->getComp() == NEIGHBORCACHE_COMP) {
			cObject** context = globalNodeList->getContext(getThisNode());
// 			if (restoreContext && context && *context) {
// 				BaseOverlayContext* overlayContext = static_cast<BaseOverlayContext*>(*context);
// 				globalNodeList->setMalicious(getThisNode(),
// 								    overlayContext->malicious);
// 				join(overlayContext->key);
// 			} else {
				join();
// 			}
		}
		delete readyMsg;
	}
	
	void SynapseController::handleTraceMessage(cMessage* msg)
	{
		delete msg;
	}
	
	void SynapseController::deliver(OverlayKey& key, cMessage* msg)
	{
		delete msg;
	}

	void SynapseController::forward(OverlayKey* key, cPacket** msg,
				NodeHandle* nextHopNode)
	{
		//TODO: LEVAMI DEBUG
		std::cout << "PACKET " << key->toString() << " FORWARDED AT NODE: " << thisNode.getKey().toString() << endl;
		delete msg;		
	}
	
	void SynapseController::update(const NodeHandle& node, bool joined)
	{
		
	}
	
	void SynapseController::forwardResponse(const OverlayKey& key, cPacket* msg,
                         const NodeHandle& nextHopNode)
	{
		delete msg;
	}
	
	// USELESS FOR NOW
	void SynapseController::registerComp(CompType compType, cModule *module) 
	{
		//register the component for the main overlay
		BaseOverlay::registerComp(compType, module);
		// then add register the component for every individual suboverlay
		// for now no TIER_N component is registered, everything goes through the SynapseController
		ConnectedOverlaysMap::iterator itOv;
		for ( itOv = connectedOverlays.begin();
			itOv != connectedOverlays.end(); ++itOv) {
			if ((compType == BOOTSTRAPLIST_COMP) ||
				(compType == NEIGHBORCACHE_COMP))
				((BaseOverlay *) itOv->second)->registerComp(compType, module);
			else 
				((BaseOverlay *) itOv->second)->registerComp(compType, this);
		}
	}
	
	void SynapseController::joinOverlay()
	{
		// Eventually do something to join the Synapse Overlay
		setOwnNodeID();
		globalNodeList->registerPeer((TransportAddress) getThisNode());
		
		// Notify all suboverlays to join
		CompReadyMessage* msg = new CompReadyMessage;
		msg->setReady(true);
		msg->setComp(CONTROLLER_COMP);
		
		ConnectedOverlaysMap::iterator itOv;
 		for (itOv = connectedOverlays.begin();
 			itOv != connectedOverlays.end(); ++itOv) {
 				sendDirect((cMessage *) msg->dup(), itOv->second->gate("direct_in"));
 		}
 		
 		delete msg;
	}

/////////////////////////////////////////////////////////////////////////////
// DISPOSABLE METHODS
/////////////////////////////////////////////////////////////////////////////
	NodeVector* SynapseController::findNode(const OverlayKey& key,
                            int numRedundantNodes,
                            int numSiblings,
                            BaseOverlayMessage* msg) 
	{
		Enter_Method("findNode()");
		std::cout << "FindNode. Mmhhh, this shouldn't be here..." << thisNode.getKey() << ", " 
				<< thisNode.getIp().str() << ":" <<thisNode.getPort() << endl;
		return new NodeVector();
	}
	
	bool SynapseController::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
                               int numSiblings, bool* err)
	{
		Enter_Method("isSiblingFor()");

		std::cout << "isSiblingFor. Mmhhh, this shouldn't be here..." << thisNode.getKey() << ", " 
				<< thisNode.getIp().str() << ":" <<thisNode.getPort() << endl;
		return false;
	}

	int SynapseController::getMaxNumSiblings()
	{
		Enter_Method("getMaxNumSiblings()");

		std::cout << "getMaximumSiblings. Mmhhh, this shouldn't be here..." << thisNode.getKey() << ", " 
				<< thisNode.getIp().str() << ":" <<thisNode.getPort() << endl;

		return false;
	}

	int SynapseController::getMaxNumRedundantNodes()
	{
		Enter_Method("getMaxNumRedundantNodes()");

		std::cout << "getMaximumRedundantNodes. Mmhhh, this shouldn't be here..." << thisNode.getKey() << ", " 
				<< thisNode.getIp().str() << ":" <<thisNode.getPort() << endl;

		return false;
	}
	
	//virtual public
	OverlayKey SynapseController::distance(const OverlayKey& x,
                                 const OverlayKey& y,
                                 bool useAlternative) const
	{
		std::cout << "distance. Mmhhh, this shouldn't be here..." << thisNode.getKey() << ", " 
				<< thisNode.getIp().str() << ":" <<thisNode.getPort() << endl;	
		return OverlayKey::UNSPECIFIED_KEY;
	}
	
	void SynapseController::setOwnNodeID()
	{
		thisNode.setKey(OverlayKey::random());
	}
}//namespace
