//
// Copyright (C) 2011 INRIA Sophia Antipolis Mediterranee
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SynapseController.h
 * @author Vincenzo Ciancaglini
 */

#ifndef __SYNAPSECONTROLLER_H_
#define __SYNAPSECONTROLLER_H_

#include "BaseOverlay.h"
#include "SynapseOverlayWrapper.h"
#include "Synapse.h"
#include "SynapseMessages_m.h"
#include "ReverseNetworkTable.h"
#include "SynapseRoutingTable.h"
#include <omnetpp.h>
#include <SHA1.h>
#include <OverlayKey.h>

//#include <CommonMessages_m.h>

namespace oversim {

/**
 * Synapse Controller Class
 *
 * 
 */
class SynapseController : public BaseOverlay
{
public:
    SynapseController();
    virtual ~SynapseController() {};
    
private:
	ReverseNetworkTable *reverseNetworkTable;
	SynapseRoutingTable *requestTable;
	
    virtual void initializeOverlay(int stage);
    virtual void finishOverlay();

    virtual void setOwnNodeID();

protected:
	// PARAMETERS
	bool synapseDebug;
	
    // reads the config file and creates a map with all the available overlays, arranged by networkID.
    void initOverlaysMap();
    // reads the config file and retrieve the connected overlay list for the node.
    void initConnectedOverlaysList();
    
    // @Override of the registerComponent, so to register the same component to in the suboverlays
    // Note! THe TIER_N components should NOT be registered in the suboverlays, at least for now...
    void registerComp(CompType compType, cModule *module);
    
    /////////////////////////////////////////////////////////////////////////////////////////////
    // FUNCTION "BORROWED" FROM BaseApp. This would need some refactoring to isolate them in   //
    // a reusable piece of code (i.e. OverlayController)                                       //
    /////////////////////////////////////////////////////////////////////////////////////////////
    
    // Override of the HandleMessage
    void handleMessage(cMessage* msg);
    void handleLowerMessage(cMessage* msg); //TODO: Change in handleOerlayMessage
    void handleUpperMessage(cMessage* msg);//TODO: Change in handleAppMessage
    void handleUDPMessage(cMessage* msg);
    void handleTCPMessage(cMessage* msg);
    void handleReadyMessage(CompReadyMessage* msg);
    void handleTraceMessage(cMessage* msg);
    void handleCommonAPILowerMessage(CommonAPIMessage* commonAPIMsg);
    void handleCommonAPIHigherMessage(CommonAPIMessage* commonAPIMsg);
    
    void handleForwardNotification(cPacket* msg, NodeHandle comp);
    bool isLegacyNode(NodeHandle node);
    // SYNAPSE METHODS
    void sendSynapseOffer(NodeHandle& dest, OverlayKey key, OverlayKey contextNetID);
    void handleSynapseOfferMessage(SynapseOfferMessage *synOffMsg);
    
    void handleSynapseRouteRequestMessage(SynapseRequestMessage *synReqMsg);
    void handleSynapseRouteResponseMessage(SynapseResponseMessage *synRespMsg);
    
    void handleSynapseRouteCall(SynapseRequestCall *synReqCall);
    void handleSynapseRouteResponse(SynapseResponseCall *synRespCall);
    void handleSynapseRouteTimeout();
    
   /**
     * Common API function: handles delivered messages from overlay
     *
     * method to handle decapsulated KBRdeliver messages from overlay module,
     * should be overwritten in derived application
     * @param key destination key
     * @param msg delivered message
     */
    void deliver(OverlayKey& key, cMessage* msg);

    /**
     * Common API function: handles messages from overlay to be forwarded
     *
     * method to handle decapsulated KBRdeliver messages from overlay module,
     * should be overwritten in derived application if needed
     * @param key destination key
     * @param msg message to forward
     * @param nextHopNode next hop
     */
    void forward(OverlayKey* key, cPacket** msg,
                         NodeHandle* nextHopNode);

    /**
     * Common API function: informs application about neighbors and own nodeID
     *
     * @param node new or lost neighbor
     * @param joined new or lost?
     */
    void update(const NodeHandle& node, bool joined);
    
    void forwardResponse(const OverlayKey& key, cPacket* msg,
                         const NodeHandle& nextHopNode);
    //////////////////////////////////////////////////////////////////////////////
    
    //
    virtual void joinOverlay();
    
    // QUESTE SARANNO DA TOGLIERE NON APPENA IL SYNAPSECONTROLLER FA IL SYNAPSECONTROLLER
    virtual NodeVector* findNode( const OverlayKey& key,
                                  int numRedundantNodes,
                                  int numSiblings,
                                  BaseOverlayMessage* msg = NULL);
    virtual bool isSiblingFor(const NodeHandle& node, const OverlayKey& key,
                               int numSiblings, bool* err);
    virtual int getMaxNumSiblings();
    virtual int getMaxNumRedundantNodes();
    virtual OverlayKey distance(const OverlayKey& x,
                                 const OverlayKey& y,
                                 bool useAlternative) const;
    // END ROBA DA TOGLIERE
						   
    // creates a new suboverlay
    cModule * createSubOverlay(const char *netID, const char *overlayName, const char *overlayType, const char *synapseModuleName);

    // Maximum number of suboverlays per node
    int maxOverlays;

    OverlaysParametersMap availableOverlaysByNetID;
    OverlaysParametersMap availableOverlaysByName;

    ConnectedOverlaysMap connectedOverlays;
    ConnectedOverlaysMap connectedOverlaysByName; //We' ll keep it here if we need it later...
};
} // namespace
#endif
