// Copyright (C) 2011 INRIA Sophia Antipolis Méditerranée
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SynapseOverlayWrapper.h
 * @author Vincenzo Ciancaglini
 */

#ifndef __SYNAPSEOVERLAYWRAPPER_H_
#define __SYNAPSEOVERLAYWRAPPER_H_

#include <OverlayKey.h>
#include "Synapse.h"
#include <CommonMessages_m.h>
#include <BaseOverlay.h>
#include "GlobalNodeList.h"
#include <GlobalNodeListAccess.h>

namespace oversim {
	
//class BootstrapList;
//class OverlayKey;

/**
	* Additional class to turn any existing overlay into a Synapse-enabled
	* overlay. Adds additional parameters such as NetID, stealth mode
	* and the additional social-based primitives.
	* 
	* To create a Synapse-enabled overlay, the class should inherit from
	* the original overlay and from SynapseBasedOverlay.
	*
	*
	* @author Vincenzo Ciancaglini
	*/
template <class BaseOverlayType=BaseOverlay>
class SynapseOverlayWrapper : public BaseOverlayType {
	
	// Constructor/destructor
public:
	SynapseOverlayWrapper()
	{
		// First, test that we have derived a BaseOverlay module
		BaseOverlay *testPointer = dynamic_cast<BaseOverlay *>(this);
		if (!testPointer)
			throw new cRuntimeError("SynapseBaseOverlay must inherit from \
			a BaseOverlay derived class!");
		
		stealthNode = false;
	};
	
	/** Virtual destructor */
	virtual ~SynapseOverlayWrapper() {};
	
	void addToGlobalList()
	{	
		GlobalNodeList *nodeList = GlobalNodeListAccess().get();
		// PeerInfo entry is basically the exact same as the SynapseController
		// No info specific to one overlay or the other...
		nodeList->addPeer(& (BaseOverlayType::getThisNode()));
	};
	
	void postInitializeOverlay()
	{
		//add the new nodehandle to the globallist
		addToGlobalList();
		
		//since the suboverlay are not the "official" overlay module
		//(that would be the synapseController) we have to override the 
		//overlay variable in order to send RPCs with the proper source node attached.
		//baseOverlayModule->setOverlay(baseOverlayModule);
		BaseOverlayType::setOverlay(this);
	};
	
protected: //fields: overlay attributes

	bool stealthNode;
	
	virtual void handleMessage(cMessage* msg)
	{
		if (dynamic_cast<CompReadyMessage*>(msg)) {
		CompReadyMessage* readyMsg = static_cast<CompReadyMessage*>(msg);
		if (readyMsg->getReady() &&
			readyMsg->getComp() == CONTROLLER_COMP) {
				BaseOverlayType::join();
		}
		delete msg;
		} else {
			BaseOverlayType::handleMessage(msg);
		}
	};

	// WE NEED TO INTERCEPT 1) OVERLAYROUTE 2) RPC. OVERLAYSIGNALING DOESN'T CARRY INFO, NEITHER APPDATA
	// HERE WE HANDLE THE OVERLAYROUTE MESSAGES, NOT THE RPC (internalHandleRpc...)
	virtual void handleBaseOverlayMessage(BaseOverlayMessage* msg,
									const OverlayKey& destKey)
	{
		//TODO: LEVAMI DEBUG
//		std::cout << "SynapseOverlayWrapper::handleBaseOverlayMessage: ";
// 		switch (msg->getType())
// 		{
// 			case OVERLAYROUTE:
// 				std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " OVERLAYROUTE" << endl;
// 				break;
// 			case APPDATA:
// 				std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " APPDATA" << endl;
// 				break;
// 			case RPC:
// 				std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " RPC" << endl;
// 				break;
// 			case OVERLAYSIGNALING:
// 				std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " OVERLAYSIGNALING" << endl;
// 				break;
// 			default:
// 				break;
// 		}
		//END DEBUG
		//1) notify the synapse controller, if a message is to be routed
		// in order to extract the netIDs AND eventually to trigger a syn_offer
		BaseRouteMessage* routeMsg = dynamic_cast<BaseRouteMessage *>(msg);
		if (routeMsg != NULL) {
			OverlayCtrlInfo* overlayCtrlInfo = check_and_cast<OverlayCtrlInfo *>(msg->getControlInfo());
			if ((!msg->isSelfMessage()) || 
				(overlayCtrlInfo->getLastHop().getIp() != BaseOverlayType::thisNode.getIp()))
				notifySynapseController(routeMsg->dup());
		}
		//2) if the message is a SYN_OFFER (routed via the overlay because in 
		// RECURSIVE_SUORCE_ROUTING), notify the controller
		// TODO: maybe at controller level...
		//3) call the original message handler
		BaseOverlayType::handleBaseOverlayMessage(msg, destKey);
	}
	
	// HERE WE HANDLE RPC CALLS NOTIFICATIONS
	virtual bool internalHandleMessage(cMessage* msg)
	{
		BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
		if (rpcMessage != NULL) {
			//TODO LEVAMI DEBUG
// 			std::cout << "SynapseOverlayWrapper::internalHandleMessage: ";
// 			switch (rpcMessage->getType())
// 			{
// 				case OVERLAYROUTE:
// 					std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " OVERLAYROUTE" << endl;
// 					break;
// 				case APPDATA:
// 					std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " APPDATA" << endl;
// 					break;
// 				case RPC:
// 					std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " RPC" << endl;
// 					break;
// 				case OVERLAYSIGNALING:
// 					std::cout << BaseOverlayType::thisNode.getIp().str() << ":" << BaseOverlayType::thisNode.getPort() << " OVERLAYSIGNALING" << endl;
// 					break;
// 				default:
// 					break;
// 			}
			
			OverlayCtrlInfo* overlayCtrlInfo = check_and_cast<OverlayCtrlInfo *>(msg->getControlInfo());
			if ((!msg->isSelfMessage()) || 
				(overlayCtrlInfo->getLastHop().getIp() != BaseOverlayType::thisNode.getIp()))
				notifySynapseController(rpcMessage->dup());
		}
		
		return BaseOverlayType::internalHandleMessage(msg);
	}
	
	void notifySynapseController(BaseOverlayMessage* msg)
	{
		KBRnotify *notifyMessage = new KBRnotify();
		notifyMessage->setType(KBR_NOTIFY);
		notifyMessage->encapsulate(msg);
		notifyMessage->setSrcCompNode(BaseOverlayType::getThisNode());
		BaseOverlayType::sendDirect(notifyMessage, BaseOverlayType::gate("appOut"));
	}

	virtual TransportAddress getBootstrapNode() 
	{
		const char *nID = BaseOverlayType::par("netID");
		BaseOverlayType::setNetworkID(OverlayKey(nID));
		return BaseOverlayType::bootstrapList->getBootstrapNode(BaseOverlayType::getNetworkID());
	}
	//------------------------------------------------------------------------
	//--- Statistics ---------------------------------------------------------
	//------------------------------------------------------------------------

private://fields: statistics

		// da aggiungere in caso...

	//------------------------------------------------------------------------
	//--- General Overlay Parameters (getter and setters) --------------------
	//------------------------------------------------------------------------

public://methods: getters and setters    
	/**
		* Returns the overlay's network ID
		**/
	bool isStealthNode()
	{ 
		return stealthNode; 
	};

	/**
		* Specifies the networkID for the overlay
		**/
	void setStealthNode(bool isStealth)
	{ 
		stealthNode = isStealth; 
	};
};
} //namsespace

#endif